"""
Copyright 2021 Daniel Afriyie

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import uuid
import stat
import errno
import shutil
from typing import Optional, Callable, Any, Sequence

from ru.hints import Path
from ru.constants import constants


def get_data(
        fn: Path,
        split: bool = False,
        split_char: Optional[str] = None,
        filter_blanks: bool = False,
) -> str | list[str]:
    """
    :param fn: filename to open
    :param split: if you want to split the data read
    :param split_char: character you want to split the data on
    :param filter_blanks: remove empty strings if split=True
    Example:
    >>>data = get_data("file.txt", split=True, split_char=",")
    >>>print(data)
    [1, 2, 3, 4]
    """
    with open(fn, encoding=constants.ENCODING) as f:
        data = f.read()
        if split:
            if split_char:
                data_split = data.split(split_char)
                if filter_blanks:
                    data_split = [s.strip() for s in data_split if s.strip() != ""]
                    return data_split
                return data_split
    return data


def mk_dir(*paths: Path) -> None:
    for p in paths:
        os.makedirs(p, exist_ok=True)


def get_filename(name: str, path: Path, is_folder: bool = False) -> Path:
    if is_folder:
        return os.path.join(path, f"{name}_{uuid.uuid4()}")

    split = name.split(".")
    stemname = ".".join(split[0:-1])
    extension = split[-1]
    return os.path.join(path, f"{stemname}_{uuid.uuid4()}.{extension}")


def handle_remove_read_only(func: Callable[[Path], None], path: Path, exc: Sequence[Any]) -> None:
    excvalue = exc[1]
    if func in (os.rmdir, os.remove) and excvalue.errno == errno.EACCES:
        os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
        func(path)
    else:
        raise


def remove_dir(p: Path, ignore_errors: Optional[bool] = True) -> None:
    try:
        shutil.rmtree(p, ignore_errors=False, onerror=handle_remove_read_only)
    except Exception as e:
        if not ignore_errors:
            raise e



"""
Copyright 2021 Daniel Afriyie

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys
import time
import warnings
import random as rd
from typing import Any, Optional, Callable, cast, overload

try:
    from selenium.webdriver.remote.webdriver import WebDriver, WebElement
    from selenium.webdriver.support import expected_conditions as ec
    from selenium.webdriver.common.action_chains import ActionChains
    from selenium.webdriver.support.wait import WebDriverWait
    from selenium.webdriver.common.by import By
    from selenium.webdriver.common.keys import Keys
    from selenium.common.exceptions import (
        TimeoutException,
        WebDriverException,
        ElementClickInterceptedException,
        StaleElementReferenceException,
    )
except ImportError:
    warnings.warn(
        "It seems you don't have selenium installed. "
        "Install it before using this module!\npip3 install selenium"
    )
    raise

try:
    import pyperclip
except ImportError:
    warnings.warn(
        "It seems you don't have pyperclip installed. "
        "Install it before using this module!\npip3 install pyperclip"
    )
    raise

type Number = int | float
type Action = Optional[str]
type Element = WebElement | list[WebElement]
type Condition[T] = Callable[[tuple[str, str]], T]


def window_scroll_to(driver: WebDriver, loc: Number) -> None:
    driver.execute_script(f"window.scrollTo(0, {loc});")


def scroll_into_view(driver: WebDriver, element: WebElement, offset: int = 200) -> None:
    window_scroll_to(driver, element.location["y"] - offset)


def scroll_into_view_js(driver: WebDriver, element: WebElement) -> None:
    driver.execute_script("arguments[0].scrollIntoView();", element)


def find_element(
        driver: WebDriver,
        locator: str,
        by: str,
        secs: Number = 10,
        condition: Condition = ec.element_to_be_clickable,
        action: Action = None,
        *args: Any,
        **kwargs: Any,
) -> Element:
    wait: WebDriverWait = WebDriverWait(driver=driver, timeout=secs)
    element: Element = wait.until(condition((by, locator)))
    if action:
        if hasattr(element, action):
            action_func = getattr(element, action)
            action_func(*args, **kwargs)
    return element


def find_element_by_xpath(
        driver: WebDriver,
        xpath: str,
        secs: Number = 10,
        condition: Condition = ec.element_to_be_clickable,
        action: Action = None,
        *args: Any,
        **kwargs: Any,
) -> Element:
    return find_element(
        driver, xpath, By.XPATH, secs, condition, action, *args, **kwargs
    )


def find_element_by_css(
        driver: WebDriver,
        selector: str,
        secs: Number = 10,
        condition: Condition = ec.element_to_be_clickable,
        action: Action = None,
        *args: Any,
        **kwargs: Any,
) -> Element:
    return find_element(
        driver, selector, By.CSS_SELECTOR, secs, condition, action, *args, **kwargs
    )


def find_element_by_id(
        driver: WebDriver,
        id: str,
        secs: Number = 10,
        condition: Condition = ec.element_to_be_clickable,
        action: Action = None,
        *args: Any,
        **kwargs: Any,
) -> Element:
    return find_element(driver, id, By.ID, secs, condition, action, *args, **kwargs)


def find_element_by_link_text(
        driver: WebDriver,
        text: str,
        secs: Number = 10,
        condition: Condition = ec.element_to_be_clickable,
        action: Action = None,
        *args: Any,
        **kwargs: Any,
) -> Element:
    return find_element(
        driver, text, By.LINK_TEXT, secs, condition, action, *args, **kwargs
    )


def driver_or_js_click(
        driver: WebDriver,
        xpath: str,
        secs: Number = 5,
        condition: Condition = ec.element_to_be_clickable,
) -> None:
    elm: WebElement
    try:
        elm = cast(WebElement, find_element_by_xpath(driver, xpath, secs=secs, condition=condition))
        ActionChains(driver).move_to_element(elm).click().perform()
    except WebDriverException:
        elm = driver.find_element(By.XPATH, xpath)
        try:
            ActionChains(driver).move_to_element(elm).click().perform()
        except WebDriverException:
            driver.execute_script("arguments[0].click()", elm)


def manual_entry(
        driver: WebDriver,
        xpath: str,
        text: str,
        secs: Number = 10,
        condition: Condition = ec.element_to_be_clickable,
        sleep_time: Number = 0.05,
        *args: Any,
) -> None:
    if (not isinstance(sleep_time, int)) and (not isinstance(sleep_time, float)):
        args += (sleep_time,)
        sleep_time = 0.05
    elm = cast(WebElement, find_element_by_xpath(driver, xpath, secs=secs, condition=condition))
    ActionChains(driver).move_to_element(elm).perform()
    elm.clear()
    text = f"{text}"
    for letter in text:
        elm.send_keys(letter)
        time.sleep(sleep_time)
    time.sleep(sleep_time)
    elm.send_keys(*args)


def enter(
        driver: WebDriver,
        xpath: str,
        text: str,
        secs: Number = 10,
        condition: Condition = ec.element_to_be_clickable,
        *args: Any,
) -> None:
    elm: WebElement = cast(WebElement, find_element_by_xpath(driver, xpath, secs=secs, condition=condition))
    ActionChains(driver).move_to_element(elm).perform()
    elm.clear()
    text = f"{text}"
    elm.send_keys(text, *args)


def paste(
        driver: WebDriver,
        xpath: str,
        text: str,
        secs: Number = 10,
        condition: Condition = ec.element_to_be_clickable,
        paste_key: str = "v",
) -> None:
    elm = cast(WebElement, find_element_by_xpath(driver, xpath, secs=secs, condition=condition))
    pyperclip.copy(text)
    ctrl = Keys.COMMAND if sys.platform == "darwin" else Keys.CONTROL
    actions = ActionChains(driver)
    actions.move_to_element(elm)
    actions.click()
    actions.key_down(ctrl)
    actions.send_keys(paste_key)
    actions.key_up(ctrl)
    actions.perform()


def random_delay(a: int = 1, b: int = 3) -> None:
    delay: int = rd.randint(a, b)
    precision: float = delay / (a + b)
    sleep_time: float = delay + precision
    time.sleep(sleep_time)


click = driver_or_js_click
driver_wait = find_element_by_xpath



"""
Copyright 2021 Daniel Afriyie

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import logging
import pathlib
import threading
import traceback
from typing import Optional
import logging.handlers as handlers
from datetime import datetime as dt

from colorama import Fore

from ru.hints import Path
from ru.constants import constants


class Logger:
    __loggers = dict[str, logging.Logger]()

    def __init__(self, name: Optional[str] = None, fmt: Optional[str] = None, filename: Optional[Path] = None) -> None:
        self._name = name or __name__
        self._fmt = fmt or "%(asctime)s:%(levelname)s:%(message)s"
        self._filename = filename
        self._root_path = pathlib.Path(".").absolute()

    def _get_log_file(self) -> Path:
        fn = self._filename or "log"
        log_dir = os.path.join(self._root_path, "logs")
        log_file = os.path.join(log_dir, f"{fn}.log")
        if not os.path.exists(log_dir):
            os.mkdir(os.path.join(log_dir))
        return log_file

    def _create_logger(self) -> logging.Logger:
        _logger = logging.getLogger(self._name)
        _logger.setLevel(logging.DEBUG)

        formatter = logging.Formatter(self._fmt)
        max_bytes = 1048576 * 100  # 100mb
        file_handler = handlers.RotatingFileHandler(
            self._get_log_file(),
            maxBytes=max_bytes,
            encoding=constants.ENCODING,
            backupCount=10,
        )
        file_handler.setFormatter(formatter)
        file_handler.setLevel(logging.DEBUG)

        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)
        stream_handler.setLevel(logging.DEBUG)

        _logger.addHandler(file_handler)
        _logger.addHandler(stream_handler)

        return _logger

    def __call__(self) -> logging.Logger:
        if self._name in self.__loggers:
            return self.__loggers[self._name]
        else:
            _logger = self._create_logger()
            self.__loggers[self._name] = _logger
            return _logger

    @staticmethod
    def get_logger(name: Optional[str] = None, fmt: Optional[str] = None, filename: Optional[Path] = None) -> logging.Logger:
        return Logger(name, fmt, filename)()


class BaseColorPrint:
    mutex: threading.Lock

    def _print(self, level: str, text: str, color: str) -> None:
        with self.mutex:
            n = dt.now().strftime("%Y-%m-%d %H:%M:%S,%f")
            t1, ms = n.split(",")
            ms = ms[0:3]
            now = f"{t1},{ms}"
            t = f"{now}:{level}:{text}"
            print(color + t + Fore.RESET)

    def info(self, text: str, color: str = Fore.CYAN) -> None:
        self._print("INFO", text, color)

    def warning(self, text: str, color: str = Fore.YELLOW) -> None:
        self._print("WARNING", text, color)

    def success(self, text: str, color: str = Fore.GREEN) -> None:
        self._print("SUCCESS", text, color)

    def error(self, error: str | BaseException, color: str = Fore.RED) -> None:
        if isinstance(error, BaseException):
            msg = self.get_error_message(error)
            self._print("ERROR", msg, color)
        else:
            self._print("ERROR", error, color)

    @staticmethod
    def get_error_message(e: BaseException) -> str:
        msg = "".join(traceback.format_exception(type(e), e, e.__traceback__))
        return msg


class ColorPrint(BaseColorPrint):
    mutex = threading.Lock()


logger = Logger.get_logger



from typing import Protocol, Callable, Any


class CastClass(Protocol):

    def __init__(self, x: str, *args: Any, **kwargs: Any) -> None: ...

type Path = str
type Cast[T] = type[CastClass] | Callable[[str], T]



"""
Copyright 2021 Daniel Afriyie

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import abc
import json
import threading
from typing import Any, Optional, override

from ru.utils import get_data
from ru.constants import constants
from ru.hints import Cast, Path, CastClass
from ru.exceptions.exceptions import ConfigKeyError, ConfigFileNotFoundError


class BaseConfig(abc.ABC):
    """ "
    Base config class for all config classes
    """

    def __init__(self, path: Path) -> None:
        self._path = path
        self._config = self.load()
        self._mutex = threading.RLock()

    @property
    def config(self) -> dict[str, str]:
        return self._config

    @abc.abstractmethod
    def save(
            self,
            filename: Path = "config.txt",
            encoding: str = constants.ENCODING,
    ) -> None:
        pass

    @abc.abstractmethod
    def load_config(self) -> dict[str, str]:
        pass

    @staticmethod
    def _cast[T](item: str, cast: Cast[T]) -> T | CastClass:
        if cast is not None:
            return cast(item.strip())
        return item

    def get[T](self, item: str, default: Any = None, cast: Optional[Cast[T]] = None) -> str | CastClass | T:
        with self._mutex:
            try:
                val = self._config[item]
                if cast:
                    return self._cast(val, cast)
                return val
            except KeyError:
                return default

    def get_as_tupple[T](self, item: str, cast: Optional[Cast[T]] = None) -> tuple[str | T | CastClass, ...]:
        with self._mutex:
            items: list[str] = self._config[item].split(",")
            if cast is None:
                return tuple(items)
            return tuple(self._cast(val, cast) for val in items)

    def load(self) -> dict[str, str]:
        try:
            return self.load_config()
        except FileNotFoundError:
            raise ConfigFileNotFoundError(
                f"{self.__class__.__name__}: Config file '{self._path}' not found!"
            )

    def __getitem__(self, item: str) -> str:
        with self._mutex:
            try:
                return self._config[item]
            except KeyError:
                raise ConfigKeyError(f"{item}")

    def __setitem__(self, key: str, value: str) -> None:
        with self._mutex:
            self._config[key] = value

    @override
    def __repr__(self) -> str:
        return str(self._config)


class JsonConfig(BaseConfig):

    def __init__(self, path: Path = "config.json") -> None:
        super().__init__(path)

    @override
    def load_config(self) -> dict[str, str]:
        with open(self._path, encoding=constants.ENCODING) as f:
            return json.load(f)

    @override
    def save(self, filename: Path = "config.txt", encoding: str = constants.ENCODING) -> None:
        with open(filename, "w", encoding=encoding) as f:
            json.dump(self._config, f, indent=4)


class TextConfig(BaseConfig):

    def __init__(self, path: Path = "config.txt"):
        super().__init__(path)

    @override
    def load_config(self) -> dict[str, str]:
        config = dict[str, str]()
        data = get_data(self._path, split=True, split_char="\n", filter_blanks=True)
        for d in data:
            try:
                split = d.split("=")
                key = split.pop(0)
                val = "=".join(split)
                config[key] = val
            except ValueError:
                pass

        return config

    @override
    def save(self, filename: Path = "config.txt", encoding: str = constants.ENCODING) -> None:
        with open(filename, "w", encoding=encoding) as f:
            text = "\n".join(
                [f"{key}={self._config[key]}" for key in self._config.keys()]
            )
            f.write(text)
